#define CL_TARGET_OPENCL_VERSION 200
#include <CL/cl_platform.h>
#include <math.h>

#include "clm.h"

#include <CL/cl.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>

typedef struct __attribute__((packed)) {
	cl_uint rows;
	cl_uint cols;
	cl_char transposed;
} cl_GPUMat;

char *loadFile(const char *path) {
	FILE *file = fopen(path, "r");
	fseek(file, 0, SEEK_END);
	size_t length = ftell(file);
	fseek(file, 0, SEEK_SET);
	char *buffer = calloc(1, length + 1);
	fread(buffer, length, 1, file);
	return buffer;
}

int main() {
	// Connect to a compute device
	int useGPU = true;
	cl_device_id deviceID;
	cl_int err = clGetDeviceIDs(NULL, useGPU ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &deviceID, NULL);
	if(err != CL_SUCCESS) {
		printf("Error: Failed to create a device group!\n");
		return 1;
	}

	char *buffer = loadFile("src/mat.cl");
	printf("%s", buffer);

	cl_context context = clCreateContext(NULL, 1, &deviceID, NULL, NULL, &err);
	if(!context) {
		printf("Failed to create context\n");
		return 1;
	}

	cl_command_queue queue = clCreateCommandQueueWithProperties(context, deviceID, NULL, &err);
	if(!queue) {
		printf("Failed to create command queue\n");
		return 1;
	}

	size_t length = strlen(buffer);
	cl_program program = clCreateProgramWithSource(context, 1, (const char **) &buffer, &length, &err);
	if(!program) {
		printf("Failed to create program\n");
		return 1;
	}

	err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to build program\n");
		// clGetProgramBuildInfo...
		return 1;
	}

	cl_kernel kernel = clCreateKernel(program, "mat_multiply", &err);
	if(!kernel) {
		printf("Failed to create kernel\n");
		return 1;
	}

	clm_Matrix a = clm_createMatrixRandom(5, 10);
	clm_Matrix b = clm_createMatrixRandom(10, 3);
	clm_Matrix out = clm_createMatrixRandom(5, 3);

	cl_GPUMat matA = {.rows = a.rows, .cols = a.cols, .transposed = a.transposed};
	cl_GPUMat matB = {.rows = b.rows, .cols = b.cols, .transposed = b.transposed};
	cl_GPUMat matOut = {.rows = out.rows, .cols = out.cols, .transposed = out.transposed};

	size_t inputSize = out.rows * out.cols;

	clm_matrixMultiplyMatrix(a, b, out);

	clm_matrixPrint(out);
	clm_matrixZero(out);

	cl_mem matA_values = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * a.rows * a.cols, NULL, &err);
	cl_mem matB_values = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * b.rows * b.cols, NULL, &err);
	cl_mem matOut_values = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(float) * out.rows * out.cols, NULL, &err);
	if(!matA_values || !matB_values || !matOut_values) {
		printf("Failed to allocate input/output buffer\n");
		return 1;
	}

	err = clEnqueueWriteBuffer(queue, matA_values, CL_TRUE, 0, sizeof(float) * a.rows * a.cols, a.values, 0, NULL, NULL);
	err |= clEnqueueWriteBuffer(queue, matB_values, CL_TRUE, 0, sizeof(float) * b.rows * b.cols, b.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to write to buffer\n");
		return 1;
	}

	err = 0;
	err |= clSetKernelArg(kernel, 0, sizeof(matA), &matA);
	err |= clSetKernelArg(kernel, 1, sizeof(matA_values), &matA_values);
	err |= clSetKernelArg(kernel, 2, sizeof(matB), &matB);
	err |= clSetKernelArg(kernel, 3, sizeof(matB_values), &matB_values);
	err |= clSetKernelArg(kernel, 4, sizeof(matOut), &matOut);
	err |= clSetKernelArg(kernel, 5, sizeof(matOut_values), &matOut_values);
	// err |= clSetKernelArg(kernel, 2, sizeof(unsigned int), &inputSize);
	if(err != CL_SUCCESS) {
		printf("Failed to set kernel args\n");
		return 1;
	}

	/*char *info = calloc(1, 1024);
	clGetProgramInfo(program, CL_PROGRAM_STRING_DEBUG_INFO, 1024, info, NULL);
	printf("INFO: %s\n", info);*/

	size_t local;
	err = clGetKernelWorkGroupInfo(kernel, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(local), &local, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to get work group size\n");
		return 1;
	}

	printf("Group size is %zu\n", local);

	size_t global = ceil((float) inputSize / local) * local;
	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global, &local, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue: %d\n", err);
		return 1;
	}

	clFinish(queue);

	err = clEnqueueReadBuffer(queue, matOut_values, CL_TRUE, 0, sizeof(float) * inputSize, out.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to read from buffer\n");
		return 1;
	}

	clm_matrixPrint(out);

	/*for(unsigned int i = 0; i < inputSize; i++) {
		if(i % 1000 != 0) continue;
		printf("%f: %f\n", inputData[i], outputData[i]);
	}*/
}
