#include "clm_gpu.h"
#include <math.h>

#define CL_TARGET_OPENCL_VERSION 200
#include <CL/cl_platform.h>

#include <CL/cl.h>

#include <stdio.h>
#include <string.h>

static cl_context context;
static cl_device_id deviceID;
static cl_command_queue queue;
static cl_kernel kernel;

struct clm_NativeBuf {
	cl_mem mem;
};

typedef struct __attribute__((packed)) {

	cl_uint rows;
	cl_uint cols;
	cl_char transposed;
} cl_GPUMat;

#define gpuMat(mat) \
	{ .rows = mat.rows, .cols = mat.cols, .transposed = mat.transposed }

char *loadFile(const char *path) {
	FILE *file = fopen(path, "r");
	fseek(file, 0, SEEK_END);
	size_t length = ftell(file);
	fseek(file, 0, SEEK_SET);
	char *buffer = calloc(1, length + 1);
	fread(buffer, length, 1, file);
	return buffer;
}

int clm_gpuInit() {
	// Connect to a compute device
	int useGPU = true;
	cl_int err = clGetDeviceIDs(NULL, useGPU ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &deviceID, NULL);
	if(err != CL_SUCCESS) {
		printf("Error: Failed to create a device group!\n");
		return 1;
	}

	char *buffer = loadFile("src/mat.cl");
	printf("%s", buffer);

	context = clCreateContext(NULL, 1, &deviceID, NULL, NULL, &err);
	if(!context) {
		printf("Failed to create context\n");
		return 1;
	}

	queue = clCreateCommandQueueWithProperties(context, deviceID, NULL, &err);
	if(!queue) {
		printf("Failed to create command queue\n");
		return 1;
	}

	size_t length = strlen(buffer);
	cl_program program = clCreateProgramWithSource(context, 1, (const char **) &buffer, &length, &err);
	if(!program) {
		printf("Failed to create program\n");
		return 1;
	}

	err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to build program\n");
		// clGetProgramBuildInfo...
		return 1;
	}

	kernel = clCreateKernel(program, "linear_forward", &err);
	if(!kernel) {
		printf("Failed to create kernel\n");
		return 1;
	}

	return 0;
}

void clm_gpuDestroy() {
}

static cl_mem allocGPUMat(cl_GPUMat mat, clm_NativeBuf *nativeBuf) {
	cl_int err;

	if(!nativeBuf->mem) {
		cl_mem mat_values = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * mat.rows * mat.cols, NULL, &err);
		if(!mat_values) {
			printf("Failed to alloc buffer: %d\n", err);
			return NULL;
		}
		nativeBuf->mem = mat_values;
	}

	return nativeBuf->mem;
}

static cl_mem writeGPUMat(cl_GPUMat gpuMat, clm_Matrix mat, clm_NativeBuf *nativeBuf) {
	cl_int err;
	cl_mem mem = allocGPUMat(gpuMat, nativeBuf);

	err = clEnqueueWriteBuffer(queue, mem, CL_TRUE, 0, sizeof(float) * mat.rows * mat.cols, mat.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue write: %d\n", err);
		return NULL;
	}

	return mem;
}

clm_NativeBuf nativeInput;

// TODO: allow writing multiple inputs at once to improve throughput (no need to rewrite weights/bias each time)
void clm_linearForward(clm_Linear *linear, clm_Matrix input) {
	if(!linear->nativeWeights) {
		linear->nativeWeights = calloc(1, sizeof(clm_NativeBuf));
		linear->nativeBias = calloc(1, sizeof(clm_NativeBuf));
		linear->nativeOutput = calloc(1, sizeof(clm_NativeBuf));
	}

	cl_GPUMat matInput = gpuMat(input);
	cl_GPUMat matWeights = gpuMat(linear->weights);
	cl_GPUMat matBias = gpuMat(linear->bias);
	cl_GPUMat matOut = gpuMat(linear->output);

	size_t workSize = matOut.rows * matOut.cols;

	cl_int err;

	cl_mem matInput_values = writeGPUMat(matInput, input, &nativeInput);
	cl_mem matWeights_values = writeGPUMat(matWeights, linear->weights, linear->nativeWeights);
	cl_mem matBias_values = writeGPUMat(matBias, linear->bias, linear->nativeBias);
	if(!matInput_values || !matWeights_values || !matBias_values) {
		linear->output = INVALID_MATRIX;
		return;
	}

	cl_mem matOut_values = allocGPUMat(matOut, linear->nativeOutput);
	if(!matOut_values) {
		printf("Failed to alloc out: %d\n", err);
		linear->output = INVALID_MATRIX;
		return;
	}

	err = 0;
	err |= clSetKernelArg(kernel, 0, sizeof(matInput), &matInput);
	err |= clSetKernelArg(kernel, 1, sizeof(matInput_values), &matInput_values);
	err |= clSetKernelArg(kernel, 2, sizeof(matWeights), &matWeights);
	err |= clSetKernelArg(kernel, 3, sizeof(matWeights_values), &matWeights_values);
	err |= clSetKernelArg(kernel, 4, sizeof(matBias), &matBias);
	err |= clSetKernelArg(kernel, 5, sizeof(matBias_values), &matBias_values);
	err |= clSetKernelArg(kernel, 6, sizeof(matOut), &matOut);
	err |= clSetKernelArg(kernel, 7, sizeof(matOut_values), &matOut_values);
	if(err != CL_SUCCESS) {
		printf("Failed to set kernel args: %d\n", err);
		return;
	}

	/*char *info = calloc(1, 1024);
	clGetProgramInfo(program, CL_PROGRAM_STRING_DEBUG_INFO, 1024, info, NULL);
	printf("INFO: %s\n", info);*/

	size_t local;
	err = clGetKernelWorkGroupInfo(kernel, deviceID, CL_KERNEL_WORK_GROUP_SIZE, sizeof(local), &local, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to get work group size\n");
		return;
	}

	size_t global = ceil((float) workSize / local) * local;
	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global, &local, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to enqueue: %d\n", err);
		return;
	}

	clFinish(queue);

	err = clEnqueueReadBuffer(queue, matOut_values, CL_TRUE, 0, sizeof(float) * workSize, linear->output.values, 0, NULL, NULL);
	if(err != CL_SUCCESS) {
		printf("Failed to read from buffer\n");
		return;
	}
}
